/**********************************************************/
/* fart - Funky Application for Rendering Terrain         */
/**********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "display.h"
#include "matrix.h"
#include "main.h"

#define TRUE 1
#define FALSE 0

/* GLOBALS FOLLOW:                      */
Display *theDisplay = (Display *) NULL;
static int      theScreen;
static int      theDepth;

static Window   theWindow = (Window) NULL;
Window   theDA = (Window) NULL;
GC       theGC;
static Colormap theColormap = (Colormap) NULL;
long     theForegroundPixel;
long     theBackgroundPixel;
int      height = 0;
int      width = 0;

Pixmap   thePixmap;

#define BORDER_WIDTH    2

void            init_display(x, y, w, h)
int             x, y, w, h;
{
   static int      first_time = TRUE;

   if (first_time)
   {
      first_time = FALSE;
      theDisplay = XOpenDisplay(NULL);
      if (theDisplay != NULL)
      {
         theScreen = DefaultScreen(theDisplay);
         theDepth = DefaultDepth(theDisplay, theScreen);
      }
      else
      {
         fprintf(stderr, "Could not open display!\n");
         exit(1);
      }
      if (create_colormap() != TRUE)
      {
         fprintf(stderr, "Error setting colormap.\n");
         exit(1);
      }
   }
   if (!open_window(x, y, w, h, "main"))
      exit(1);
}

void            close_display()
{
   if (!close_window())
      exit(1);
}


/****************************************************************/
/* Function:    open_window                                                                             */
/* Parameters:  int     x       X Position of window                                    */
/*                              int     y       Y Position of window                                    */
/*                              int     h       Height of window                                                */
/*                              int     w       Width of window                                                 */
/*                              char    *name   Name of the window                              */
/****************************************************************/

int             open_window(x, y, w, h, name)
int             x, y, w, h;
char           *name;
{
   XSetWindowAttributes theWindowAttributes;
   XSizeHints      theSizeHints;
   unsigned long   theWindowMask;
   XGCValues       theGCValues;

   theForegroundPixel = WhitePixel(theDisplay, theScreen);
   theBackgroundPixel = BlackPixel(theDisplay, theScreen);

   width = w;
   height = h;

   theWindowAttributes.border_pixel = BlackPixel(theDisplay, theScreen);
   theWindowAttributes.background_pixel = theBackgroundPixel;
/* theWindowAttributes.background_pixel= theForegroundPixel; */
   theWindowAttributes.backing_store = Always;
   theWindowMask = CWBackPixel | CWBorderPixel | CWBackingStore;

   theWindow = XCreateWindow(theDisplay,
      RootWindow(theDisplay, theScreen),
      x, y, width, height, BORDER_WIDTH,
      theDepth, InputOutput, CopyFromParent,
      theWindowMask, &theWindowAttributes);
   XStoreName(theDisplay, theWindow, name);

   theSizeHints.flags = USPosition | USSize | PMinSize | PMaxSize;
   theSizeHints.x = x;
   theSizeHints.y = y;
   theSizeHints.width = width;
   theSizeHints.height = height;
   theSizeHints.min_width = width;
   theSizeHints.max_width = width;
   theSizeHints.min_height = height;
   theSizeHints.max_height = height;
   XSetNormalHints(theDisplay, theWindow, &theSizeHints);

   theDA = XCreateSimpleWindow(theDisplay, theWindow,
      0, 0, width, height, 0, CopyFromParent, theBackgroundPixel);
   thePixmap = XCreatePixmap(theDisplay, theDA, width, height, theDepth);

   /* Now creat the Graphics Context     */
   theGC = XCreateGC(theDisplay, thePixmap, (unsigned long) 0, &theGCValues);
   if (theGC == 0)
   {
      fprintf(stderr, "Could not create GC!\n");
      return (FALSE);
   }
   XSetForeground(theDisplay, theGC, theForegroundPixel);
   XSetBackground(theDisplay, theGC, theBackgroundPixel);

   XMapWindow(theDisplay, theWindow);
   XMapSubwindows(theDisplay, theWindow);
   XFlush(theDisplay);
   return (TRUE);
}

int             close_window()
{
   if (theWindow != (Window) NULL)
   {
      XFreeGC(theDisplay, theGC);
      XDestroyWindow(theDisplay, theWindow);
      theWindow = (Window) NULL;
   }
   if (theColormap != (Colormap) NULL)
   {
      XFreeColormap(theDisplay, theColormap);
      theColormap = (Colormap) NULL;
   }
   return (TRUE);
}

int             create_colormap()
{
   int             loop1;
   XColor          color_cells[256];

   theColormap = XCreateColormap(theDisplay, RootWindow(theDisplay,
         theScreen), DefaultVisual(theDisplay, theScreen),
      AllocAll);

   for (loop1 = 0; loop1 < 256; ++loop1)
   {
      color_cells[loop1].flags = DoRed | DoBlue | DoGreen;
      color_cells[loop1].pad = 0;
      color_cells[loop1].pixel = (long) loop1;
   }
   for (loop1 = 0; loop1 < 256; ++loop1)
   {
      color_cells[loop1].red = loop1 * 256;
      color_cells[loop1].green = 0;
      color_cells[loop1].blue = 0;
   }
   XStoreColors(theDisplay, theColormap, color_cells, 256);
   return (TRUE);
}


/* NOTE: !!!!!!!! pixmap == theWindow */

void            draw_point(p)
point          *p;
{
   vector          v;

   XSetForeground(theDisplay, theGC, theForegroundPixel);
   point2vector(v, p);
   if (pipe_point(v))
      XDrawPoint(theDisplay, thePixmap, theGC, (int) v[0], (int) v[1]);
}

void            draw_points(plist, n)
point           plist[];
int             n;
{
   XPoint         *xp;
   vector          v;
   int             i, count;

   XSetForeground(theDisplay, theGC, theForegroundPixel);
   xp = (XPoint *) malloc(sizeof(XPoint) * n);
   if (xp == 0)
   {
      fprintf(stderr, "Error allocating points.\n");
      return;
   }
   count = 0;
   for (i = 0; i < n; i++)
   {
      point2vector(v, &plist[i]);
      if (pipe_point(v))
      {
         xp[count].x = (short) v[0];
         xp[count].y = (short) v[1];
         count++;
      }
   }
   if (count)
      XDrawPoints(theDisplay, thePixmap, theGC, xp, count, CoordModeOrigin);
   free(xp);
}

void            draw_vector(v1, v2, col)
vector          v1, v2;
int             col;
{
   XSetForeground(theDisplay, theGC, theForegroundPixel);
   if (pipe_vector(v1, v2))
   {
      XDrawLine(theDisplay, thePixmap, theGC,
         (int) v1[0], (int) v1[1], (int) v2[0], (int) v2[1]);
   }
}

void            draw_polyline(ar, n, col)
point           ar[];
int             n, col;
{
   vector          v1, v2;
   int             i;

   XSetForeground(theDisplay, theGC, theForegroundPixel);
   for (i = 0; i < n - 1; i++)
   {
      point2vector(v1, &ar[i]);
      point2vector(v2, &ar[i + 1]);
      if (pipe_vector(v1, v2))
      {
         XDrawLine(theDisplay, thePixmap, theGC,
            (int) v1[0], (int) v1[1], (int) v2[0], (int) v2[1]);
      }
   }
   point2vector(v1, &ar[i]);
   point2vector(v2, &ar[0]);
   if (pipe_vector(v1, v2))
   {
      XDrawLine(theDisplay, thePixmap, theGC,
         (int) v1[0], (int) v1[1], (int) v2[0], (int) v2[1]);
   }
}

void            flush()
{
   XFlush(theDisplay);
}

void            clear_window()
{
   XSetForeground(theDisplay, theGC, theBackgroundPixel);
   XFillRectangle(theDisplay, thePixmap, theGC, 0, 0, width, height);
   XCopyArea(theDisplay, thePixmap, theDA, theGC, 0, 0, width, height, 0, 0);
   XFlush(theDisplay);
}

void            show_window()
{
   XCopyArea(theDisplay, thePixmap, theDA, theGC, 0, 0, width, height, 0, 0);
   XFlush(theDisplay);
}

short           get_pointer(x, y, mask)
int            *x, *y;
unsigned int   *mask;
{
   Window          w1, w2;
   int             rx, ry, wx, wy;
   unsigned int    m;

   if (XQueryPointer(theDisplay, theWindow,
         &w1, &w2, &rx, &ry, &wx, &wy, &m))
   {
      if (w2 == theDA)
      {
         *x = wx;
         *y = wy;
         *mask = m;
         return (1);
      }
   }
   return (0);
}
